#!/usr/bin/env python3
import os
import json
from sysrepo_api import sysrepo_api3 as api
from netaddr import *
from oslo_utils import netutils


class SysrepoDhcpRelay(api.SysrepoAPI):

    def __init__(self,
                 app_name="dhcprelay",
                 log_file=None,
                 log_level=None,
                 console_log=False,
                 namespace="/ucpe:config/dhcprelay:dhcp-relay",
                 run_as_subscriber=False,
                 priority=0):
        api.SysrepoAPI.__init__(
            self,
            app_name=app_name,
            log_file=log_file,
            log_level=log_level,
            console_log=console_log,
            namespace=namespace,
            run_as_subscriber=run_as_subscriber,
            priority=priority)

    def is_ipv4(self, ipaddr):
        """Check the IP address version and return True in case of IPv4"""
        if ipaddr is not None and ipaddr != "":
            first_addr = ipaddr.split(',')[0]
            return netutils.is_valid_ipv4(first_addr)

        return False

    def get_system_bind_iface_address(self, iface=""):
        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(iface))
        if address != '' and address is not None:
            ip = IPNetwork(address)
            return str(ip.ip)
        else:
            return None

    def get_system_bind_iface_netmask(self, iface=""):
        address = self.get_item("/ucpe:config/interfaces:interface[name='{}']/ipv4-prefix".format(iface))
        if address != '' and address is not None:
            ip = IPNetwork(address)
            return str(ip.netmask)
        else:
            return None

    def apply_dhcprelay2system(self, relay_srv, relay_int, link_int):
        if relay_srv is None:
            return False
        if relay_int is None:
            return False
        
        if 'data' in relay_int:
            relay_int = 'br-{}'.format(relay_int)
        
        with open('/etc/default/isc-dhcp-relay', 'w') as cfg_file:
            cfg_file.write("# Defaults for isc-dhcp-relay initscript\n# sourced by /etc/init.d/isc-dhcp-relay\n")
            cfg_file.write("# installed at /etc/default/isc-dhcp-relay by the maintainer scripts\n\n")
            cfg_file.write("#\n# This is a POSIX shell fragment\n#\n\n")
            cfg_file.write("# What servers should the DHCP relay forward requests to?\n")
            cfg_file.write("SERVERS=\"" + relay_srv.replace(',', ' ') + "\"\n\n")
            cfg_file.write("# On what interfaces should the DHCP relay (dhrelay) serve DHCP requests?\n")
            cfg_file.write("INTERFACES=\"" + relay_int.replace(',', ' ') + "\"\n\n")
            cfg_file.write("# Additional options that are passed to the DHCP relay daemon?\n")
            if link_int:
                cfg_file.write("OPTIONS= \"-U " + link_int + "\"\n")
            else:
                cfg_file.write("OPTIONS=\"\"\n")

            cfg_file.flush()
            os.fsync(cfg_file.fileno())

        return True

    def apply_dhcp6relay2system(self, relay_srv, relay_int, link_int):
        if relay_srv is None:
            return False
        if relay_int is None:
            return False

        if 'data' in relay_int:
            relay_int = 'br-{}'.format(relay_int)

        with open('/etc/default/isc-dhcp-relay6', 'w') as cfg_file:
            cfg_file.write("# Defaults for isc-dhcp-relay6 initscript\n# sourced by /etc/init.d/isc-dhcp-relay6\n")
            cfg_file.write("# installed at /etc/default/isc-dhcp-relay6 by the maintainer scripts\n\n")
            cfg_file.write("#\n# This is a POSIX shell fragment\n#\n\n")
            cfg_file.write("# What servers should the DHCP relay forward requests to?\n")
            if link_int:
                cfg_file.write("SERVERS=\""
                               + relay_srv.replace(',', '%{} '.format(link_int))
                               + "%{}".format(link_int) + "\"\n\n")
            else:
                cfg_file.write("SERVERS=\"" + relay_srv.replace(',', ' ') + "\"\n\n")

            cfg_file.write("# On what interfaces should the DHCP relay (dhrelay) serve DHCP requests?\n")
            cfg_file.write("INTERFACES=\"" + relay_int.replace(',', ' ') + "\"\n\n")
            cfg_file.flush()
            os.fsync(cfg_file.fileno())

        return True

    def start_dhcp_service(self, service):
        if service == "isc-dhcp-relay.service":
            self.stop_dhcp_service("isc-dhcp-relay6.service")
        else:
            self.stop_dhcp_service("isc-dhcp-relay.service")

        self.execute_shell_command("systemctl enable {}".format(service))
        self.execute_shell_command("systemctl --no-block start {}".format(service))

    def stop_dhcp_service(self, service):
        self.execute_shell_command("systemctl --no-block stop {}".format(service))
        self.execute_shell_command("systemctl reset-failed {}  || true".format(service))
        self.execute_shell_command("systemctl disable {}".format(service))

    def apply_startup(self):
        self.logger.debug('Apply STARTUP config for DHCP Relay')
        sys_conf = self.get_startup_config(full_config=False, full_path=False)
        if sys_conf is None:
            sys_conf = {}

        self.logger.debug('Startup config : %s' % json.dumps(sys_conf))
        relay_state = False
        relay_srv = ''
        relay_int = ''
        link_int = ''

        if 'enabled' in sys_conf:
            if sys_conf['enabled'] and 'true' in str(sys_conf['enabled']).lower():
                relay_state = True
            else:
                relay_state = False
        if 'dhcp-relay-server' in sys_conf:
            relay_srv = sys_conf['dhcp-relay-server']
        if 'dhcp-relay-interface' in sys_conf:
            relay_int = sys_conf['dhcp-relay-interface']
        if 'link-selection-sub-option' in sys_conf:
            link_int = sys_conf['link-selection-sub-option']

        if relay_state:
            if self.is_ipv4(relay_srv):
                dhcp_service = "isc-dhcp-relay.service"
                res = self.apply_dhcprelay2system(relay_srv, relay_int, link_int)
            else:
                dhcp_service = "isc-dhcp-relay6.service"
                res = self.apply_dhcp6relay2system(relay_srv, relay_int, link_int)

            if res is False:
                self.logger.error('Failed creating DHCP relay configuration file')
                return True
            self.start_dhcp_service(dhcp_service)
        else:
            self.apply_dhcprelay2system("", "", "")
            self.apply_dhcp6relay2system("", "", "")
            self.stop_dhcp_service("isc-dhcp-relay.service")
            self.stop_dhcp_service("isc-dhcp-relay6.service")

        return True

    def apply_running(self, change_set):
        self.logger.debug('Apply RUNNING config for DHCP Relay')
        sys_conf = self.get_running_config(full_path=False)
        relay_state = False
        relay_srv = ''
        relay_int = ''
        link_int = ''

        if 'enabled' in sys_conf:
            if sys_conf['enabled'] and 'true' in str(sys_conf['enabled']).lower():
                relay_state = True
            else:
                relay_state = False
        if 'dhcp-relay-server' in sys_conf:
            relay_srv = sys_conf['dhcp-relay-server']
        if 'dhcp-relay-interface' in sys_conf:
            relay_int = sys_conf['dhcp-relay-interface']
        if 'link-selection-sub-option' in sys_conf:
            link_int = sys_conf['link-selection-sub-option']

        for change in change_set:
            if not change["xpath"].startswith(self.namespace):
                continue
            if not 'operation' in change:
                self.logger.error('Malformed change set')
                continue

            self.logger.debug('Operation -->> %s' % change['operation'])
            self.logger.debug('XPath -->> %s' % change['xpath'])

            if change['operation'] == 'MODIFIED':
                if "dhcp-relay/enabled" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' %
                                      (change['xpath'], change['old_value'], change['new_value']))
                    if change['new_value'] and 'true' in str(change['new_value']).lower():
                        relay_state = True
                    else:
                        relay_state = False
                if "dhcp-relay-server" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' %
                                      (change['xpath'], change['old_value'], change['new_value']))
                    relay_srv = change['new_value']
                if "dhcp-relay-interface" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' %
                                      (change['xpath'], change['old_value'], change['new_value']))
                    relay_int = change['new_value']
                if "link-selection-sub-option" in change["xpath"]:
                    self.logger.debug('Modified  - %s: %s -> %s' %
                                      (change['xpath'], change['old_value'], change['new_value']))
                    link_int = change['new_value']
            elif change['operation'] == 'MOVED':
                self.logger.debug('Moved  - %s' % change['xpath'])
            elif change['operation'] == 'CREATED':
                self.logger.debug('Created - %s' % change['xpath'])
                if "dhcp-relay-server" in change["xpath"]:
                    self.logger.debug('Created  - %s: -> %s' % (change['xpath'], change['new_value']))
                    relay_srv = change['new_value']
                if "dhcp-relay-interface" in change["xpath"]:
                    self.logger.debug('Created  - %s: -> %s' % (change['xpath'], change['new_value']))
                    relay_int = change['new_value']
                if "link-selection-sub-option" in change["xpath"]:
                    self.logger.debug('Created  - %s: -> %s' % (change['xpath'], change['new_value']))
                    link_int = change['new_value']
            elif change['operation'] == 'DELETED':
                self.logger.debug('Deleted - %s, \'%s\'' % (change['xpath'], change['old_value']))
                if "dhcp-relay-server" in change["xpath"]:
                    self.logger.debug('Deleted  - %s: -> %s' % (change['xpath'], change['old_value']))
                    relay_srv = ''
                if "dhcp-relay-interface" in change["xpath"]:
                    self.logger.debug('Deleted  - %s: -> %s' % (change['xpath'], change['old_value']))
                    relay_int = ''
                if "link-selection-sub-option" in change["xpath"]:
                    self.logger.debug('Deleted  - %s: -> %s' % (change['xpath'], change['old_value']))
                    link_int = ''
            else:
                self.logger.error('Unknown operation \'%s\'' % change['operation'])

        if relay_state:
            if self.is_ipv4(relay_srv):
                dhcp_service = "isc-dhcp-relay.service"
                res = self.apply_dhcprelay2system(relay_srv, relay_int, link_int)
            else:
                dhcp_service = "isc-dhcp-relay6.service"
                res = self.apply_dhcp6relay2system(relay_srv, relay_int, link_int)

            if res is False:
                self.logger.error('Failed creating DHCP relay configuration file')
                return True
            self.start_dhcp_service(dhcp_service)
        else:
            self.apply_dhcprelay2system("", "", "")
            self.apply_dhcp6relay2system("", "", "")
            self.stop_dhcp_service("isc-dhcp-relay.service")
            self.stop_dhcp_service("isc-dhcp-relay6.service")

        return True

    def apply_config(self, change_set=None, apply_startup=False):
        ret = False
        if apply_startup:
            ret = self.apply_startup()
        else:
            ret = self.apply_running(change_set)

        return ret

    def set_state(self, state):
        self.set_item("{}/enabled".format(self.namespace), state)

    def get_state(self):
        return self.get_item("{}/enabled".format(self.namespace))

    def unset_state(self):
        self.delete_item("{}/enabled".format(self.namespace))

    def set_relay_server(self, relay_srv):
        self.set_item("{}/dhcp-relay-server".format(self.namespace), relay_srv)

    def get_relay_server(self):
        return self.get_item("{}/dhcp-relay-server".format(self.namespace))

    def unset_relay_server(self):
        self.delete_item("{}/dhcp-relay-server".format(self.namespace))

    def set_relay_int(self, relay_int):
        self.set_item("{}/dhcp-relay-interface".format(self.namespace), relay_int)

    def get_relay_int(self):
        return self.get_item("{}/dhcp-relay-interface".format(self.namespace))

    def unset_relay_int(self):
        self.delete_item("{}/dhcp-relay-interface".format(self.namespace))

    def set_relay_link_int(self, link_int):
        self.set_item("{}/link-selection-sub-option".format(self.namespace), link_int)

    def get_relay_link_int(self):
        return self.get_item("{}/link-selection-sub-option".format(self.namespace))

    def unset_relay_link_int(self):
        self.delete_item("{}/link-selection-sub-option".format(self.namespace))
